#! /usr/bin/env python


# spaces.py
# This program simulates a number of proposed models of word
# segmentation. It works by reading a transcription of speech, and
# processing it into proposed input representations (phonemes, syllables,
# etc). It then runs a suite of segmentation modules on each of the
# input representations. The performance of each module is then
# analyzed.
# 
#
# Preprocessing:
# 
# The program reads in a phonetic transcription of speech. The notation
# used to describe phonemes is arbitrary, and the markings used in the
# transcription to to indicate utterance, word, syllable, and phoneme
# boundaries, and stress informaiton are strings which can be arbitrarily 
# designated in the main execution script of the program. 
# 
# 
# Segmentation:
# 
# Broadly, three types of information are used for segmentation:
# information which is thought to be available in any superficial
# representation of a speech stream (utterance boundaries and the
# boundaries between whatever perceptual building blocks make up the
# speech stream (phonemes, syllables, or max-onset syllables in the case
# of this program)), stress information, and statistical information (in
# the form of transitional probability between the perceptual building
# blocks of the speech stream). Segmentation modules use this information to 
# segment the speech stream.
# 
# 
# Performance Analysis:
# 
# Performance is reported as precision and recall of spaces at the
# phoneme level. That is, after each phoneme, there is a potential
# space. Recall is calculated as the number of those potential spaces
# which are correctly guessed divided by the total number of spaces in
# the document. Precision is calculated as the number of correct
# guesses, divided by the total number of guesses made by each segmentation 
# module



# Support Functions
#
# These functions are called throughout the program

def strip(input_list, string_to_strip):

	# remove a "string_to_strip" from each item in "input_list"
	
	# compile a regular expression to do the stripping
	stripper = re.compile(string_to_strip)
	
	# initialize an empty output list
	stripped_list = []
	
	# strip "string_to_strip" from each item in "input_list" and put the 
	# leftovers in "stripped_list"
	for item in input_list:
		stripped_list.append(stripper.sub('', item))
	return stripped_list
	

def performance(guess_list, answer_key):

	# compare a guess_list and and answer_key, calculating precision and recall

	# initialize counters at zero
	hits = 0
	misses = 0
	false_alarms = 0
	
	# compare each item in the guess_list and answer_key, and keep track of the
	# the number of times they agreed on a space (hits), the guess_list missed a
	# space (misses), or a space was guessed that wasn't actually there 	
	# (false_alarms).
	for i in range (len(answer_key)):
		if guess_list[i] == 1 and answer_key[i] == 1:
			hits += 1
		elif guess_list[i] == 0 and answer_key[i] == 1:
			misses += 1
		elif guess_list[i] == 1 and answer_key[i] == 0:
			false_alarms += 1

	# calculate recall and precision (and make sure python knows that we're
	# dealing with floating point numbers)
	recall = float(float(hits)/float(hits+misses))
	precision = float(float(hits)/float(hits + false_alarms))
	print "   recall:", recall
	print "   precision:", precision
	
	
# Preprocessing Functions
#
# These functions are called by the big preprocessing function
# "preprocess(speech_file)"

def syllabify(phoneme_string):
	
	# make a list of all onsets used in the speech file by extracting
	# all string which occur between an utterance boundary and a vowel
 	onset_pattern = utterance_delimiter + '(.*?)[aeiou]'
 	all_onsets = re.findall(onset_pattern, phoneme_string)
 	
 	# remove duplicates from the list of onsets
 	unsorted_valid_onsets = []
 	for onset in all_onsets:
 		if onset not in unsorted_valid_onsets:
 			unsorted_valid_onsets.append(onset)
 	
 	
 	# sort the valid onsets by length (longest first)
 	
 	# initialize an empty list to keep track of lengths in
 	lengths = []
 	
 	# for each item in "unsorted_valid_onsets", put a tuple in
 	# "lengths" of the format (number of phonemes in onset, index of
 	# onset in "unsorted_valid_onsets")
 	for i in range(len(unsorted_valid_onsets)):
 		lengths.append( (unsorted_valid_onsets[i].count(phoneme_delimiter), i) )
 	
 	# sort the items in "lengths" by the number of phonemes in the onset
 	lengths.sort()
 	
 	# initialize a new list to put the onsets into, sorted by length
 	valid_onsets = []
 	
 	# put onsets into the new list "valid_onsets", shortest to longest,
 	# by looking at the indexes in the tuples in "lengths" 
 	for length_of_onset, i in lengths:
 		valid_onsets.append(unsorted_valid_onsets[i])
 	
 	# reverse "valid_onsets" so that onsets are in order from longest to 
	# shortest
 	valid_onsets.reverse()
 	
 	# convert "valid_onsets" into a list that the regex module can understand,
 	# that is, as a string of items separated by "|" 
 	onsets = string.join(valid_onsets, "|")
 	
 	
 	# make a list of all the vowels used in the speech file
 	all_vowels = re.findall('[aeiou].*?' + phoneme_delimiter, phoneme_string)
 	
 	# remove duplicates from the list of vowels
 	vowel_list = []
 	for vowel in all_vowels:
 		if vowel not in vowel_list:
 			vowel_list.append(vowel)
 			
 	# convert "vowel_list" into a list that the regex module can understand		
 	vowels = string.join(vowel_list, "|")
 	
 	
 	# find syllables by searching for non-overlapping instances of the format
 	# (coda, utterance boundary marker (no match is also), onset, nucleus).
 	# The longest onset is always preferred because the list of onsets is 
	# ordered by length (longest first) and the regex module goes with the
	# first match if it has a list of possible matches.
 	syllable_pattern = '(.*?)(' + utterance_delimiter + '*)(' + onsets + ')(' + vowels +')'
	syllables = re.findall(syllable_pattern, phoneme_string)

	# clean up syllables (rearrange the results of the regex findall into the 
	# same syllable format used in the rest of the program)
	syllable_list = []
	mark = ''
	onset = ''
	nucleus = ''
	for i in range (len(syllables) - 1):
		coda = syllables[i][0]
		syllable = mark + onset + nucleus + coda
		syllable_list.append(syllable)
		mark = syllables[i][1]
		onset = syllables[i][2]
		nucleus = syllables[i][3]
		
	# drop the first syllable from the list (it's just got initialization values 
	# in it) 
	syllable_list = syllable_list[1:]

	return syllable_list


# Segementation Functions
#
# These functions are called by the big segmentation function
# segment(speech_list)

def silences(speech_list):

	# guess a space at every utterance boundary
	
	print 'Segmenting at utterance boundaries'
		
	# initialize the guess_list as a list of the length of the speech file in 
	# phonemes (getting the length of the answer_key is a shortcut to make 
	# things run faster; no information is taken from the answer_key that would 
	# effect segmentation performance.)
	guess_list = [0] * len(answer_key)
	
	#start at phoneme zero
	phoneme_number = 0
	
	# if there is an utterance delimiter in an item of the speech list, put a
	# space before it
	for item in speech_list:
		if utterance_delimiter in item:
			guess_list[phoneme_number] = 1
		phoneme_number += item.count(phoneme_delimiter)
	
	# print out precision and recall
	performance (guess_list, answer_key)
	
	
def every_unit(speech_list):

	# guess a space before each item in the speech list
	
	print 'Segmenting after all ' + representation
	
	# initialize the guess_list as a list of the length of the speech file in 
	# phonemes (getting the length of the answer_key is a shortcut to make 
	# things run faster; no information is taken from the answer_key that would 
	# effect segmentation performance.)
	guess_list = [0] * len(answer_key)
	
	phoneme_number = 0
	
	for item in speech_list:
		guess_list[phoneme_number] = 1
		phoneme_number += item.count(phoneme_delimiter)
	
	performance(guess_list, answer_key)


def stress_initial(speech_list):

	print 'Segementing before primary stressed ' + representation
	
	guess_list = [0] * len(answer_key)
	phoneme_number = 0
	for item in speech_list:
		if primary_stress_marker in item:
			guess_list[phoneme_number] = 1
		phoneme_number += item.count(phoneme_delimiter)
	
	performance(guess_list, answer_key)


def stress_local(speech_list):

	print 'Segmenting between pairs of primary stressed ' + representation
	
	guess_list = [0] * len(answer_key)
	phoneme_number = 0
	this_item = ''
	for next_item in speech_list:
		if ((primary_stress_marker in this_item) and (primary_stress_marker in next_item)):
			guess_list[phoneme_number] = 1
		phoneme_number += next_item.count(phoneme_delimiter)
		this_item = next_item
	performance(guess_list, answer_key)
	

def calculate_tps(speech_list):

	# set up some variables for use in the counter
	pair_frequencies = {}
	item_counter = {}
	this_item = ''

	# count items in the speech list	
	for next_item in speech_list:

		# count the number of times each item occurs
		if item_counter.has_key(this_item):
			item_counter[this_item] += 1
		else:
			item_counter[this_item] = 1

		# count the number of times each pair occurs
		if pair_frequencies.has_key((this_item, next_item)):
			pair_frequencies[this_item, next_item] += 1
		else:
			pair_frequencies[this_item, next_item] = 1

		this_item = next_item
	
	
	# set up a hash table for transitional probabilities
	tps = {}
	
	# calculate transitional probabilities from counted items
	for pair in pair_frequencies.keys():
	
		pair_frequency = float(pair_frequencies[pair])
		item_frequency = float(item_counter[pair[0]])
		
		tps[pair] = float(pair_frequency/item_frequency)
	
	return tps

def average_dictionary_values(dictionary):
	#calculate the average of all the values in a dictionary (hash table)
	
	# initialize counters
	total = 0
	counter = 0
	
	# for each key in the dictionary, get the associated value
	for key in dictionary.keys():
		total += dictionary[key]
		counter += 1
	# average the values if the dictionary had anything in it	
	if counter > 0:
		average = float(float(total) / float(counter))
	#if the dictionary didn't have any values in it, return the average 0
	else:
		average = 0
	
	# return the average
	return average

def tp_local_min(speech_list, tps):

	print 'Segmenting at local minima in tp between ' + representation
	
	# set up the guess list
	guess_list = [0] * len(answer_key)
	
	# set up phoneme_number, to keep track of where we are in the speech list
	phoneme_number = 0
	
	# so we only have to get 1 tp at once, initialize the first 2 tp values to 0
	this_tp = 0
	last_tp = 0
	
	# initialize last_item with the first item in the speech list
	last_item = speech_list[0]
	
	# go through the speech list, checking to see if we're at a local min in tp
	for item in speech_list[1:]:
	
		#get the next transitional probability (3 needed to calculate local min)
		next_tp = tps[(last_item, item)]

		# guess a space if we're at a local min in tp
		if last_tp > this_tp < next_tp:
			guess_list[phoneme_number] = 1
		
		# move the transitional probabilities back one slot
		last_tp = this_tp
		this_tp = next_tp

		# update the number of the phoneme that we're on
		phoneme_number += last_item.count(phoneme_delimiter)

		# move the current item back into the last_item slot
		last_item = item
	
	# calculate performance
	performance(guess_list, answer_key)	
	

def tp_threshold(speech_list, threshold, tps):
	
	print 'Segmenting where tp between ' + representation + ' < ' + str(threshold)
	guess_list = [0]*len(answer_key)
	
	phoneme_number = 0
	
	# initialize last_item with the first item in the speech list
	last_item = speech_list[0]
	
	for item in speech_list[1:]:
		
		tp = tps[(last_item, item)]
		if tp < threshold:
			guess_list[phoneme_number] = 1
		phoneme_number += last_item.count(phoneme_delimiter)
		
		last_item = item

	performance(guess_list, answer_key)
	


# Main Functions
#
# These functions are called by the main execution script for the
# program

def preprocess(speech_file):

	# initialize a hash table
	# for different representations of the speech stream
	speech = {}

	# read the speech file
	input_file = open(speech_file, 'r')
	input_string = input_file.read()

	# split the speech file into lists of phonemes and syllables	
	marked_phonemes = re.split(phoneme_delimiter, input_string)
	
	# replace phoneme delimiters that were stripped by re.split
	for i in range(len(marked_phonemes)):
		marked_phonemes[i] = marked_phonemes[i] + phoneme_delimiter
		
	speech['marked phonemes'] = marked_phonemes
	marked_syllables = re.split(syllable_delimiter, input_string)

	# remove word and syllable delimiters
	word_and_syllable_delim = '[' + syllable_delimiter + word_delimiter + ']'
	
	phonemes = strip(marked_phonemes, word_and_syllable_delim)
	perceptual_syllables = strip(marked_syllables, word_and_syllable_delim)
	
	speech['phonemes'] = phonemes
	speech['perceptual syllables'] = perceptual_syllables
	
	# build syllables using the max-onset principle
	phoneme_string = re.sub(word_and_syllable_delim, '', input_string)
	
	max_onset_syllables = syllabify(phoneme_string)
	
	speech['max onset syllables'] = max_onset_syllables
	
	# return the useful stuff
	return speech


def generate_answer_key(marked_speech_list):

	# generate an answer key based on a speech list with word markers at word 
	# boundaries
	
	#initialize the answer key as a list of zeros
	answer_key = [0] * len(marked_speech_list)

	# if there is a word delimiter in the speech list, put a 1 at that location 
	# in the answer key
	for i in range(len(marked_speech_list)):
		if word_delimiter in marked_speech_list[i]:
			answer_key[i] = 1	
	return answer_key


def segment(speech_list):
	
	# segment using surface features of the speech stream
	silences(speech_list)
	every_unit(speech_list)
	
	# segment using stress information
	stress_initial(speech_list)
	stress_local(speech_list)
	
	# remove stress information from the speech stream
	speech_list = strip(speech_list, '[' + stress_marker + utterance_delimiter + ']')
	
	# calculate TPs in advance, to make things run faster
	tps = calculate_tps(speech_list)
		
	# segment using statistical learning
	tp_local_min(speech_list, tps)
	
	# choose a threshold below for the tp threshold segmentation method

	threshold = average_dictionary_values(tps)	# <- thresh calc'd as average tp

	#threshold = .50 							# <- arbitrary threshold value

	#for threshold in range(21)[1:]:			# <- threshold = [0.0:0.5:1.0]
	#	threshold = .05 * threshold
	#	tp_threshold(speech_list, threshold, tps)
	
	tp_threshold(speech_list, threshold, tps)

	
	
# Main Program

import re
import string

speech_file = 'mother.speech.txt'

# variables for markers used in the speech file
utterance_delimiter = 'U'
word_delimiter = 'W'
syllable_delimiter = 'S'
phoneme_delimiter = 'P'
stress_marker = '0-9'
primary_stress_marker = '1'

print
print "Preprocessing speech file \"" + speech_file + "\""
speech = preprocess(speech_file)

print "Generating answer key"
answer_key = generate_answer_key(speech['marked phonemes'])

for representation in ['phonemes']:
	print "Running word segmentation using", representation, "as input representation"
	segment(speech[representation])
	print
